import time

import socketio

from F import DATE
from F.CLASS import FairClass
from FW.FairSocket.Message import FairMessage
import F


"""
    -> For Connecting to a Server.
    -> One Way Communication to Server.
        - This Client is not opened up and waiting for outside messages.
"""

class FairClient(FairClass):
    serverUrl = 'http://localhost:3671'
    clientId = F.get_uuid()
    socket = socketio.Client()
    userName = "FairClient"
    messages = {}
    callback = False
    eventName = "onMessage"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.serverUrl:
            self.serverUrl = input("Server Url -> ")
        self.event_binder()
        # firstInput = input("Who do you want to message?")

    def connect(self):
        self.socket.connect(self.serverUrl)

    def event_binder(self):
        funcs = self.get_method_names(self)
        for func in funcs:
            if str(func).startswith("on"):
                eventName = str(func)
                eventFunc = self.get_func(func)
                self.register_event(eventName, eventFunc)
        self.register_user_event()

    def register_event(self, eventName, eventFunction):
        self.socket.on(eventName)(eventFunction)

    def register_user_event(self):
        if self.userName:
            eventFunc = self.get_func("onMessage")
            self.register_event(self.userName, eventFunc)

    def emitOnConnect(self):
        self.socket.emit("onConnect", { "fromEventName": self.eventName, "userName": self.userName })

    def emit(self, eventName:str, eventMessage):
        self.socket.emit(eventName, eventMessage)

    def onMessage(self, data):
        fm = FairMessage().fromJson(data)
        if not fm:
            print(f"-> {data} ")
        else:
            if fm.userName == self.userName:
                return
            self.addToMessages(fm)
            print(f"{fm.userName}-> {fm.message}")
        self._sendCallback(data=data)

    def addToMessages(self, fairMessage):
        dateTime = DATE.get_log_date_time_dt()
        self.messages[dateTime] = fairMessage

    def onResponse(self, data):
        print(data)  # {'from': 'server'}

    def _sendCallback(self, data):
        if self.callback:
            self.callback(data)

def startWebSocketClient(serverUrl=None):
    socket = FairClient(serverUrl=serverUrl)
    socket.connect()
    return socket


if __name__ == '__main__':
    sock = FairClient(userName="terminal")
    messObj = FairMessage()
    sock.connect()
    sock.emitOnConnect()
    sock.emit("onGetUsers", {"fromEventName": "onMessage"})
    while True:
        time.sleep(1)
        mess = input("\n$ ")
        fm = FairMessage(message=mess, userName="terminal")
        sock.emit("onMessage", fm.toJson())
        time.sleep(1)


from F import LIST

urls = [
    "https://towardsdatascience.com/how-to-use-qgis-spatial-algorithms-with-python-scripts-4bf980e39898",
"https://www.cnn.com/2022/03/27/politics/joe-biden-vladimir-putin-ukraine-war/index.html",
"https://towardsdatascience.com/a-step-by-step-guide-to-scheduling-tasks-for-your-data-science-project-d7df4531fc41",
    "https://www.forbes.com/sites/investor/2022/01/28/ethereum-cardano-polygon-solana-avalanche-and-polkadot-deathmatch-who-wins-who-dies/?sh=4292a4897c9b",
"https://finance.yahoo.com/news/metaverse-real-estate-market-growing-115600231.html",
"https://cointelegraph.com/news/price-analysis-1-28-btc-eth-bnb-ada-sol-xrp-luna-doge-dot-avax",
"https://www.forbes.com/sites/investor/2022/01/28/ethereum-cardano-polygon-solana-avalanche-and-polkadot-deathmatch-who-wins-who-dies/?sh=4292a4897c9b",
"https://www.businessinsider.com/tomi-lahren-spoke-at-police-training-event-dismissed-reform-wapo-2022-1",
"https://www.foxbusiness.com/business-leaders/elon-musk-offers-support-to-canadian-truckers-amid-covid-vaccine-mandate",
"https://towardsdatascience.com/a-step-by-step-guide-to-scheduling-tasks-for-your-data-science-project-d7df4531fc41",
"https://www.americanbanker.com/payments/news/inside-ripples-plans-for-mainstream-crypto-payments",
"https://finance.yahoo.com/news/metaverse-real-estate-market-growing-115600231.html",
"https://cointelegraph.com/news/blockchain-metaverse-ecosystems-gain-traction-as-brands-create-digital-experiences",
"https://www.theguardian.com/world/2022/jan/26/biden-threatens-putin-with-personal-sanctions-if-russia-invades-ukraine",
"https://www.americanbanker.com/payments/news/inside-ripples-plans-for-mainstream-crypto-payments",
"https://towardsdatascience.com/decorators-in-python-fundamentals-for-data-scientists-eada7f4eba85",
"https://www.wsj.com/articles/meta-platforms-facebook-fb-q1-earnings-report-2022-1165102219",
"https://www.wsj.com/articles/primary-elections-2022-south-carolina-nevada-races-test-trumps-sway-in-gop-11655199002",
"https://finance.yahoo.com/news/metaverse-real-estate-market-growing-115600231.html",
"https://www.newsobserver.com/news/business/article260136540.html",
"https://www.reuters.com/business/media-telecom/jury-alex-jones-defamation-case-begin-deliberations-punitive-damages-2022-08-05/"

]

import HttpRequest

test1OnlySuccess = []
test2OnlySuccess = []
bothSuccess = []
bothFailed = []

def loop_urls():
    for url in urls:
        test1 = HttpRequest.get_request_v2(url)
        test1Result = LIST.get(0, test1, default=False)
        test2 = HttpRequest.get_request_urllib(url)
        test2Result = LIST.get(0, test2, default=False)
        if test1Result and test2Result:
            bothSuccess.append(url)
        elif test1Result:
            test1OnlySuccess.append(url)
        elif test2Result:
            test2OnlySuccess.append(url)
        else:
            bothFailed.append(url)


loop_urls()
print(test1OnlySuccess)


import time
import F
from F import OS
from F.CLASS import Thread, FairClass
from FW.FairSocket import Client, Server

"""
    -> Two Way Communication between Server/Clients.
        - This Server is opened up and waiting for outside messages.
"""
ngrokToken = ""

class FairSocket(FairClass):
    socketManagerId = F.get_uuid()
    socketManagerName = f"FairServer:{OS.get_username()}"
    server = None
    client = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.server = Server.FairServer()
        Thread.runFuncInBackground(Server.FairServer().start)
        # self.server.start()
        print("Yep, we're still running.")
        time.sleep(5)
        # Client.startWebSocketClient(serverUrl="http://localhost:3671")
        self.client = Client.FairClient(serverUrl="http://localhost:3671")
        self.client.connect()
        self.client.emit("onPrinter", "What up!!!!!!!!")
        tester = input("Give me something: ")
        self.client.emit("onPrinter", tester)



if __name__ == '__main__':
    FairSocket()


from flask import Flask
from flask_socketio import SocketIO, join_room, leave_room, emit, send
from pyngrok import ngrok
from pyngrok.ngrok import NgrokTunnel

import F
from F import OS
from F.CLASS import FairClass
from FW.FairSocket.Message import FairMessage


"""
    -> Two Way Communication between Server/Clients.
        - This Server is opened up and waiting for outside messages.
"""
ngrokToken = ""

class FairServer(FairClass):
    serverID = F.get_uuid()
    serverName = f"FairServer:{OS.get_username()}"
    app = None
    socket = None
    host = '0.0.0.0'
    port = 3671
    tunnel = ngrok
    ngrokToken = None
    publicURL: NgrokTunnel = ""
    #->
    clients = []
    subscribed_users = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = Flask(__name__)
        self.socket = SocketIO(self.app)
        if self.get_arg("makePublic", kwargs, default=False):
            self.make_public()
        self.event_binder()

    def make_public(self):
        if not self.ngrokToken:
            return
        self.tunnel.set_auth_token(self.ngrokToken)
        self.publicURL = self.tunnel.connect(self.port, "http")
        print(f"\n{self.publicURL}\n")

    def event_binder(self):
        funcs = self.get_method_names(self)
        for func in funcs:
            if str(func).startswith("on"):
                eventName = str(func)
                eventFunc = self.get_func(func)
                self.register_event(eventName, eventFunc)

    def register_event(self, eventName, eventFunction):
        self.socket.on(eventName)(eventFunction)

    def start(self):
        self.socket.run(app=self.app, host=self.host, port=self.port, debug=False)

    def onConnect(self, data):
        print("ON CONNECTED!!!", data, self)
        user = self.get_arg("userName", data, default=False)
        fromEventName = self.get_arg("fromEventName", data)
        if fromEventName:
            self.clients.append(fromEventName)
        self.emit('onResponse', f"Welcome to the Game. {user}")

    def onStartProcess(self, data):
        pass

    def onStopProcess(self, data):
        pass

    def onMessage(self, data):
        messageObj = FairMessage().fromJson(data)
        print(f" IN -> {messageObj.message}")
        sent = []
        for c in self.clients:
            if c in sent:
                continue
            sent.append(c)
            self.sendFairMessage(c, messageObj)

    def onGetUsers(self, data):
        fromEventName = self.get_arg("fromEventName", data, default=False)
        self.emit(fromEventName, self.clients)

    def onPrinter(self, data):
        print(f"Printer-> {data} ")

    def emit(self, eventName:str, eventMessage:{}):
        self.socket.emit(eventName, eventMessage)

    def sendFairMessage(self, eventName:str, fairMessage:FairMessage):
        self.socket.emit("onMessage", fairMessage.toJson())

def startWebSocketServer():
    socket = FairServer()
    socket.start()
    return socket


if __name__ == '__main__':
    socket = FairServer(initNgrok=True)
    socket.start()


import os

from FW.Crawler import UrlProcess
from FW.Crawler import UrlExtractor
from F import DICT
from FNLP import URL
from FW.Core import HttpRequest
from FW.Crawler.UrlQueue import FQueue
from F.LOG import Log
Log = Log("FWEB.Crawler.ArchiveCrawler_v2")

way_back_machine_url = "https://web.archive.org"
avoid_list = ['youtube', 'twitter', 'facebook', 'instagram', 'advertising', 'apple', 'google',
              'spotify', 'tiktok', 'soundcloud', 'linkedin', 'flickr', 'oauth', 'terms-of-service',
              'privacy-policy', 'contact-us', '.fr', '.de', "coupon.", "coupons.", ".coupon", ".coupons",
              ".chinese", "chinese.", ".japan", "japan.", ".kr", ".fr", ".getinvisiblehand"]

MAX_QUEUE = 100

CRAWLER_VERSION = "3.0.0"

class SoupError(Exception):
    def __init__(self, message, errors=None):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        # Now for your custom code...
        self.errors = errors

class KillThySelf(Exception):
    def __init__(self, message, errors=None):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        # Now for your custom code...
        self.errors = errors

class Modes:
    NORMAL = "NORMAL"
    SUICIDE = "SUICIDE"
    KILL = "KILL"

class ArchiveCrawler:
    pid = None
    timer = None
    original_url = None
    current_base_url = None
    # -> kwargs
    max = MAX_QUEUE
    suicideSwitch = False
    saveToDB = True
    # -> Global
    MODE = Modes.NORMAL
    queue: FQueue = None
    extraction_success = []
    extraction_failed = []
    stay_within = ""
    soup = None
    total_count = 0
    base_site = ""
    # -> Alternative Output
    articles = []

    def __init__(self, url, **kwargs):
        """
        -> Takes single URL, grabs every URL, then tries to download them all.
        -> Takes next URL, grabs every URL, tries to download them all.
        -> !Potential for Forever-Loop!
        """
        self.handle_kwargs(**kwargs)
        self.pid = os.getpid()
        self.original_url = url
        self.stay_within = URL.get_site_name(url)
        self.base_site = URL.get_base_url(url)
        self.current_base_url = self.base_site
        self.queue = FQueue(maxSize=self.max, avoidList=avoid_list)
        Log.i(f"Launching ArchiveCrawler. URL=[ {url} ] PID=[ {self.pid} ]")
        self.queue.add(url)
        Log.i(f"Ready to START: ArchiveCrawler PID=[ {self.pid} ]")

    """
            -> GLOBAL CRAWLER INIT & RUN HELPERS
        1. Input URL and everything else is pre-setup.
    """

    @classmethod
    def start_UnlimitedMode(cls, _url, **kwargs):
        kwargs["suicideMode"] = False
        newClass = cls(_url, **kwargs)
        newClass.run()
        return newClass

    @classmethod
    def start_SuicideMode(cls, _url, **kwargs):
        kwargs["suicideMode"] = True
        newClass = cls(_url, **kwargs)
        newClass.run()
        return newClass

    """
        -> INIT HELPERS
    1. Set Global Configurations.
        - 'max' = Max number of URL's allowed into queue.
        - 'suicideMode' = Once 'max' is reached, crawler will empty queue and end/die.
    2. Run Main Queue Loop Process.
    """

    def handle_kwargs(self, **kwargs):
        self.max = DICT.get("max", kwargs, default=MAX_QUEUE)
        self.suicideSwitch = DICT.get("suicideMode", kwargs, default=False)
        self.saveToDB = DICT.get("saveToDB", kwargs, default=True)

    def run(self):
        Log.i(f"STARTING: ArchiveCrawler PID=[ {self.pid} ]")
        self.run_spider_queue()

    """
        -> MAIN CRAWLER PROCESS/LOOP
    1. Loop Main Queue until Empty.
    2. Process/Download URL into Article.
    3. Extract all further URL's from current URL's HTML.
    """

    """ (Crawler.UrlQueue.py) -> 1. Start: Run Queue until Empty """
    def run_spider_queue(self):
        Log.i(f"Starting URL Queue with {self.queue.size()}.")
        while not self.queue.isEmpty():
            self.total_count += 1
            # -> Clean Queue Check.
            if self.queue.size() > self.max / 2:
                Log.v("Cleaning Queue.")
                self.queue.clean()
            Log.v("Popping URL from Main Queue.")
            _url = self.queue.get()
            Log.i(f"Current URL = [ {_url} ]")
            if not self.stay_within:
                self.current_base_url = URL.get_base_url(_url)
            self.url_process(_url)
            # -> Begin to die? or Continue as Normal?
            if self.verify_mode():
                # -> Run URL Extraction
                self.url_extractor(_url)
            # -> Print Current Status for Developer
            self.print_status()
        # -> The End.
        return self

    """ (Crawler.UrlProcess.py) -> 2. Process URL """
    def url_process(self, _url):
        # -> Attempt Download
        result = UrlProcess.attempt_download(_url, self.saveToDB)
        if result:
            if not self.saveToDB:
                self.articles.append(result)
            self.extraction_success.append(_url)
        else:
            self.extraction_failed.append(_url)

    """ (Crawler.UrlExtractor.py) -> 3. Extract all URLs from current URLs WebPage. """
    def url_extractor(self, _url):
        # -> Request HTML and Parse into Soup Object.
        self.soup = HttpRequest.get_request_3k_to_html(_url, parseToSoup=True)
        if self.soup is None:
            return False
        Log.v("Extracting URLs via Url Extractor.")
        if self.soup.soup:
            extracted_urls = UrlExtractor.extract_urls_from_soup(self.soup.soup, self.current_base_url, stayWithin=self.stay_within)
        else:
            extracted_urls = UrlExtractor.extract_urls_from_soup(self.soup, self.current_base_url, stayWithin=self.stay_within)
        # -> Add all URLs to Queue
        if extracted_urls:
            self.queue.add_list(extracted_urls)
        return True

    """
    -> HELPERS
    """

    def verify_mode(self):
        """ Check Mode. """
        if self.MODE == Modes.NORMAL:
            self.suicideCheck()
            return True
        if self.MODE == Modes.SUICIDE:
            return False
        if self.MODE == Modes.KILL:
            # -> If time goes beyond X amount in the future...
            self.killThySelf()

    def change_mode(self, newMode):
        self.MODE = newMode

    def suicideCheck(self):
        """
        If suicideMode,
        check if crawler needs to switch
        to turning off the queue and emptying queue
        or not.
        """
        if self.suicideSwitch:
            queue_count = self.queue.size()
            if self.queue.isFull() \
                    or queue_count >= self.max - 1\
                    or self.total_count >= self.max:
                Log.i("Queue is FULL! Changing MODE to SUICIDE.")
                self.change_mode(Modes.SUICIDE)

    def killThySelf(self):
        """ Immediately cause a fetal error and end system process. """
        raise KillThySelf("Kill Order Initiated.")

    """
    -> Print/Log Status for Developer
    """
    def print_status(self):
        Log.i(f"---------------------------------------------")
        # Log.i(f"Time: {DATE.to_hours_minutes_seconds(self.timer.current_time())} Hours:Minutes:Seconds.")
        Log.i(f"Crawler Mode: {self.MODE}")
        Log.i(f"Spider Queue Size: {self.queue.size()}")
        Log.i(f"Total URLs Crawled: {self.total_count}")
        Log.i(f"Article Success: {len(self.extraction_success)}")
        Log.i(f"Article Failed: {len(self.extraction_failed)}")
        Log.i(f"---------------------------------------------")






from FW.FairSocket.Message import FairMessage
# from Server import FairServer
# from Client import FairClient
from pyngrok import ngrok

def openNgrokPort(port, ngrokToken):
    if not ngrokToken:
        return
    tunnel = ngrok
    tunnel.set_auth_token(ngrokToken)
    publicURL = tunnel.connect(port, "http")
    print(f"\n{publicURL}\n")
    return tunnel


# def getWebSocketServer():
#     return FairServer()
#
# def startWebSocketServer():
#     socket = FairServer()
#     socket.start()
#     return socket


# def getWebSocketClient(serverUrl=None):
#     return FairClient(serverUrl=serverUrl)
#
# def startWebSocketClient(serverUrl=None):
#     socket = FairClient(serverUrl=serverUrl)
#     socket.connect()
#     return socket
#
# sock = FairClient(userName="chace")
# messObj = FairMessage()
# sock.connect()
# messObj.message = "you are gay?"
# messObj.senderEventName = "onMessage"
# sock.emit("onConnect", messObj.toJson())
# sock.emit("onGetUsers", {"fromEventName": "chace"})


import F
from F import DATE
from F.CLASS import FairClass

class FairMessage(FairClass):
    messageId = F.get_uuid()
    date = DATE.get_now_month_day_year_str()
    userName = ""
    sender = ""
    fromEventName = "onMessage"
    receiver = ""
    toEventName = "onMessage"
    message = ""
    action = ""



from FW.Crawler.Crawl import ArchiveCrawler

DEFAULT_MAX_QUEUE = 500

def run_ListMode(url, maxQueue=DEFAULT_MAX_QUEUE):
    c = ArchiveCrawler.start_SuicideMode(_url=url, max=maxQueue, saveToDB=False)
    return c

def run_suicideMode(url, maxQueue=DEFAULT_MAX_QUEUE):
    c = ArchiveCrawler.start_SuicideMode(_url=url, max=maxQueue)
    return c

def run_unlimitedMode(url, maxQueue=DEFAULT_MAX_QUEUE):
    c = ArchiveCrawler.start_UnlimitedMode(_url=url, maxQueue=maxQueue)
    return c

if __name__ == '__main__':
    url = "https://web.archive.org/web/20130921072250/http://www.bloomberg.com/technology/"
    url2 = "https://web.archive.org/web/20131021173221/http://www.buzzfeed.com/"
    url3 = "https://www.nbcnews.com/"
    url_medium = "https://medium.com/"
    medium = "https://medium.com/"
    science = "https://www.science.org/"
    coindesk = 'https://www.coindesk.com/'
    cryptnews = "https://cryptonews.com/"
    coinbase = "https://blog.coinbase.com/"
    tokenEconomy = "https://tokeneconomy.co/archive"
    bbc = "https://www.bbc.com/"
    guardian = "https://www.theguardian.com/world/2022/jan/26/biden-threatens-putin-with-personal-sanctions-if-russia-invades-ukraine"
    marketwatch = "https://www.marketwatch.com/"
    imdb = "https://www.imdb.com/title/tt0110357/news?ref_=tt_ql_sm"
    cnet = "https://www.cnet.com/news/"
    verge = "https://www.theverge.com/tech"
    yahoo = "https://finance.yahoo.com/"
    engadget = "https://www.engadget.com/"
    meta1 = "https://www.investors.com"
    meta2 = "https://www.minergate.com/blog"
    meta3 = "https://www.pocketgamer.biz"
    meta4 = "https://www.uktech.news"
    meta5 = "https://www.newsanyway.com"
    meta6 = "https://cointelegraph.com/tags/nft"
    meta7 = "https://cryptoslate.com/defi/"
    meta8 = "https://medium.com/tag/machine-learning"
    meta9 = "https://www.barrons.com/topics/technology?mod=BOL_TOPNAV"
    meta10 = "http://www.coinnewsasia.com/"
    # if guard in guardian:
    #     print(True)
    run_suicideMode(meta10)


from F import DICT
from F import DATE
from FW.JParser import Keys
from F.LOG import Log
from FW.Core import Extractor

Log = Log("FairArticle.Parser.JsonParser")

""" import keys() method from Keys Module. """
keys = Keys.keys


def parse(data, parseAll=False, client=False, extractDate=False) -> {}:
    Log.d("Parsing data IN", v=f"=[ {data} ]")
    try:
        json_obj = {}
        # Core All
        json_obj["author"] = DICT.get_all_keys(data, keys("author"), force_type=True)
        json_obj["title"] = DICT.get_all_keys(data, keys("title"), force_type=True)
        json_obj["body"] = DICT.get_all_keys(data, keys("body"), force_type=True)
        json_obj["url"] = DICT.get_all_keys(data, keys("url"), force_type=True)
        json_obj["source_url"] = DICT.get_all_keys(data, keys("source_url"), force_type=True)
        json_obj["client"] = client
        json_obj["date_created"] = DATE.parse_obj_to_month_day_year_str(DATE.get_now_date_dt())
        json_obj["tags"] = DICT.get_all_keys(data, keys("tags"))
        json_obj["description"] = DICT.get_all_keys(data, keys("description"), force_type=True)
        json_obj["img_url"] = DICT.get_all_keys(data, keys("imgUrl"), force_type=True)
        # Date Extraction Attempt (might be None/False)
        published_date = DICT.get_all_keys(data, keys("published_date"))
        # If Date
        if published_date:
            pub_date = DATE.parse_obj_to_month_day_year_str(published_date)
        else:
            # If no date, but extractDate setting is enabled.
            if extractDate:
                rawHTML = DICT.get("html", data, False)
                extractedDate = Extractor.ExtractDateFromHTML(rawHTML)
                if extractedDate:
                    pub_date = extractedDate
                else:
                    pub_date = DATE.parse_obj_to_month_day_year_str(DATE.get_now_date_dt())
            # No Date, plus no extractDate setting.
            else:
                pub_date = DATE.parse_obj_to_month_day_year_str(DATE.get_now_date_dt())
        json_obj["published_date"] = pub_date
        json_obj["pub_date"] = DATE.TO_DATETIME(pub_date)  # for mongodb queries. -> probably my biggest fuck up so far...
        # TODO: -> Grab a list of images to use.
        # json_obj["img_urls"] = DICT.get_all_keys(data, keys("imgUrl"), force_type=True)
        # enhanced -> Jarticle
        if parseAll:
            json_obj["tickers"] = DICT.get_all_keys(data, keys("tickers"))
            json_obj["keywords"] = DICT.get_all_keys(data, keys("keywords"))
            json_obj["summary"] = DICT.get("summary", data)
            json_obj["comments"] = DICT.get("comments", data)
            json_obj["source_rank"] = DICT.get("source_rank", data)
            json_obj["category"] = DICT.get("category", data)
            json_obj["sentiment"] = DICT.get("sentiment", data)
            json_obj["category_scores"] = DICT.get("category_scores", data)
            json_obj["score"] = DICT.get("score", data)

        Log.v(f"Parsing data OUT=[ {json_obj} ]")
        return json_obj
    except Exception as e:
        Log.e(f"Failed to parse data=[ {data} ]", e)
        return None


def parse_from_reddit_client(raw_post):
    json_objc = {}
    author = DICT.get("author", raw_post)
    # Core
    json_objc["author"] = DICT.get("name", author)
    json_objc["title"] = DICT.get("title", raw_post)
    json_objc["body"] = DICT.get("selftext", raw_post, default="Post-Body-Empty")
    json_objc["url"] = DICT.get("url", raw_post)
    dt = DATE.parse_reddit_timestamp_to_datetime(DICT.get("created_utc", raw_post))
    json_objc["published_date"] = DATE.parse_obj_to_month_day_year_str(dt)
    json_objc["date_created"] = DATE.parse_obj_to_month_day_year_str(DATE.get_now_date_dt())
    json_objc["client"] = "reddit client"
    json_objc["source"] = "reddit"
    json_objc["source_url"] = "www.reddit.com"
    # + Reddit Core
    tempSub = DICT.get("subreddit", raw_post)
    json_objc["subreddit"] = f"r/{str(tempSub)}"
    json_objc["up_votes"] = DICT.get("ups", raw_post)
    json_objc["upvote_ratio"] = DICT.get("upvote_ratio", raw_post)
    json_objc["post_score"] = DICT.get("score", raw_post)
    json_objc["comments"] = parse_comments_into_json_list(raw_post.comments)
    json_objc["comment_count"] = DICT.get("num_comments", raw_post)
    return json_objc


def parse_comments_into_json_list(comments):
    temp_list = []
    for comment in comments:
        if len(temp_list) > 200:
            continue
        try:
            user_name = comment.author.name
            if user_name == "AutoModerator":
                continue
        except Exception as e:
            Log.d("No User for Comment", e)
            continue
        dt = DATE.parse_reddit_timestamp_to_datetime(comment.created_utc)
        temp_json = {
            "author": user_name,
            "body": comment.body,
            "up_votes": comment.ups,
            "published_date": DATE.parse_obj_to_month_day_year_str(dt)
        }
        temp_list.append(temp_json)
    return temp_list

def parse_from_twitter_client(raw_tweet):
    json_objc = {}
    # Core
    json_objc["author"] = DICT.get("name", raw_tweet)
    json_objc["username"] = DICT.get("screen_name", raw_tweet)
    json_objc["body"] = DICT.get("full_text", raw_tweet, default="Tweet-Body-Empty")
    urls = DICT.get("urls", raw_tweet)
    url = False
    if urls:
        for item in urls:
            url = DICT.get("expanded_url", item, default=False)
            if url:
                break
    if not url:
        url = DICT.get("url", raw_tweet)
    json_objc["url"] = url
    json_objc["published_date"] = DATE.parse_obj_to_month_day_year_str(DICT.get("created_at", raw_tweet))
    json_objc["date_created"] = DATE.parse_obj_to_month_day_year_str(DATE.get_now_date_dt())
    json_objc["client"] = "twitter client"
    json_objc["source"] = "twitter"
    json_objc["source_url"] = "www.twitter.com"
    # + Twitter Core
    json_objc["retweet_count"] = DICT.get("retweet_count", raw_tweet)
    json_objc["verified"] = DICT.get("verified", raw_tweet)
    return json_objc


from newspaper import Article
from FW.JParser import JsonParser
from FNLP import URL
from F.LOG import Log
Log = Log("FairWEB.newspaper3k.ArticleDownloader")

def download_article(url):
    """
    - Attempts to download HTML/URL
    - Converts/Returns obj to HookupObj
    - Optional ability to validate hookupObj before returning.
    """
    article = Article(url)
    try:
        # -> Download HTML from URL
        article.download()
        # -> Initial Parsing from HTML to Article Object
        article.parse()
        article.nlp()
        Log.s(f"Successfully Downloaded Article. Attempting to Parse into Hookup.")
        # -> Parse from Article Object into Hookup Object
        json = JsonParser.parse(article, client="newspaper3k", extractDate=True)
        json["source"] = URL.get_site_name(url)
        return json
    except Exception as e:
        Log.e("Failed to Download Article.", error=e)
        return False

def download_html(url) -> bool:
    """
    - Attempts to download HTML/URL
    - Converts/Returns obj to HookupObj
    - Optional ability to validate hookupObj before returning.
    """
    article = Article(url)
    try:
        # -> Download HTML from URL
        article.download()
        html = article.html
        Log.s(f"Successfully Downloaded Article. Attempting to Parse into Hookup.")
        if html:
            return html
        return False
    except Exception as e:
        Log.e("Failed to Download Article.", error=e)
        return False

def parse_html(url, html):
    article = Article(url)
    article.html = html
    article.parse()
    article.nlp()
    # -> Parse from Article Object into Hookup Object
    json = JsonParser.parse(article, client="newspaper3k", extractDate=True)
    json["source"] = URL.get_site_name(url)
    return json



import requests

def get_external_ip():
    """Get your external IP address as string.
    Uses httpbin(1): HTTP Request & Response Service
    """
    return requests.get("http://httpbin.org/ip").json().get('origin')

def duckdns_update(domains, token, newIp=None, verbose=False):
    """Update duckdns.org Dynamic DNS record.
    Args:
        domains (str): The DuckDNS domains to update as comma separated list.
        token (str): An UUID4 provided by DuckDNS for your user.
        verbose (bool): Returns info about whether or not IP has been changed as
            well as if the request was accepted.
    Returns:
        "OK" or "KO" depending on success or failure. Verbose adds IP and change
        status as well.
    """
    params = {
        "domains": domains,
        "token": token,
        "ip": newIp if not None else get_external_ip(),
        "verbose": verbose
    }
    r = requests.get("https://www.duckdns.org/update", params)
    return r.text.strip().replace('\n', ' ')



from F import DICT

from FNLP import URL
from FNLP.Language import Utils
from FNLP.Regex import Re, ReDate
from F import LIST
from F import DATE
from FW.Core import HttpRequest, Tag, Soup
from F.LOG import Log
from dateutil import parser
import sys

Log = Log("Core.Extractor")

DATE_TAGS = ['rnews:datePublished', 'article:published_time', 'OriginalPublicationDate',
             'datePublished', 'og:published_time', 'article_date_original',
             'publication_date', 'sailthru.date', 'PublishDate', 'pubdate',
             'og:published_time', 'og:pubdate', 'published_time', 'date']

DESCRIPTION_TAGS = ["og:description", "description"]

NO_STRINGS = set()
A_REL_TAG_SELECTOR = "a[rel=tag]"
A_HREF_TAG_SELECTOR = ("a[href*='/tag/'], a[href*='/tags/'], "
                       "a[href*='/topic/'], a[href*='?keyword=']")
RE_LANG = r'^[A-Za-z]{2}$'

good_paths = ['story', 'article', 'feature', 'featured', 'slides',
              'slideshow', 'gallery', 'news', 'video', 'media',
              'v', 'radio', 'press']
bad_chunks = ['careers', 'contact', 'about', 'faq', 'terms', 'privacy',
              'advert', 'preferences', 'feedback', 'info', 'browse', 'howto',
              'account', 'subscribe', 'donate', 'shop', 'admin']
bad_domains = ['amazon', 'doubleclick', 'twitter']

s = " "
UNKNOWN = "UNKNOWN"
# -> Master Parser
def parse_str(obj: str):
    return parser.parse(obj)

def parse_date(obj=None):
    try:
        if type(obj) is str:
            obj = parse_str(obj)
        elif type(obj) is list:
            return None
        p_date = str(obj.strftime("%B")) + s + str(obj.strftime("%d")) + s + str(obj.strftime("%Y"))
        return p_date
    except Exception as e:
        print(e)
        return False

items = ["author", "source", "url",
         "source_url", "img_url", "title",
         "date", "description", "body", "keywords"]

def ExtractDate(url):
    """ Convenience Method """
    return Extractor.Extract_PublishedDate(url)

def ExtractDateFromHTML(RawHTML):
    """ Convenience Method """
    try:
        return Extractor.Extract_PublishedDateFromRawHTML(RawHTML=RawHTML)
    except Exception as e:
        Log.e("Failed to extract publish date.", error=e)
        return False

class Extractor:
    base_url = ""
    isReddit = False
    subReddit = None
    soup = None
    data = {}

    @classmethod
    def Extract(cls, soup, url, client=False):
        sys.setrecursionlimit(10000)
        newCls = cls()
        newCls.base_url = URL.extract_base_url(url)
        newCls.set_data("client", client)
        if Re.contains("reddit", url):
            newCls.isReddit = True
            newCls.subReddit = URL.extract_sub_reddit(url)
            newCls.set_data("subreddit", newCls.subReddit)
        newCls.soup = soup
        newCls.build_json()
        return newCls

    @classmethod
    def Extract_PublishedDate(cls, url):
        sys.setrecursionlimit(10000)
        newCls = cls()
        # soup = HttpRequest.request_to_html(url)
        soup2 = HttpRequest.get_request_3k_to_html(url)
        newCls.soup = soup2
        newCls.base_url = URL.extract_base_url(url)
        if Re.contains("reddit", url):
            newCls.isReddit = True
            newCls.subReddit = URL.extract_sub_reddit(url)
            newCls.set_data("subreddit", newCls.subReddit)
        if newCls.date():
            return newCls.data["published_date"]
        return False

    @classmethod
    def Extract_PublishedDateFromRawHTML(cls, RawHTML):
        if not RawHTML:
            return False
        sys.setrecursionlimit(10000)
        newCls = cls()
        # soup = HttpRequest.request_to_html(url)
        newCls.soup = Soup.Parse(rawText=RawHTML)
        if newCls.date():
            date1 = DICT.get("published_date", newCls.data, False)
            if date1:
                return date1
            else:
                return DICT.get("date", newCls.data, False)
        return False

    def build_json(self):
        """  DYNAMIC {JSON/DICT} BUILDER  """
        self.data = {}
        for item in items:
            func = self.safe_get_att(item)
            if func:
                try:
                    func()
                except Exception as e:
                    Log.e(f"Failed to extract: {func}", error=e)

    def set_data(self, key, value):
        self.data[key] = value

    def safe_get_att(self, attr):
        try:
            item = getattr(self, attr)
            return item
        except Exception as e:
            Log.e("Failed to get attribute/function.", error=e)
            return False

    @staticmethod
    def get_content(search_results):
        temp_tag = LIST.get(2, search_results)
        temp_value = Tag.get_value_for_key(temp_tag, "content")
        return temp_value

    """
        -> Get Attributes
    """
    def keywords(self):
        test_tag = self.soup.soup.findAll("meta")
        for item in test_tag:
            test = Tag.get_attribute(item, "name")
            if test and Re.contains_any(["keywords"], test):
                keywords = Tag.get_attribute(item, "content")
                self.set_data("keywords", keywords)
                return True
        return False

    # -> Date <- #
    def date(self):
        try:
            if self.isReddit:
                self.reddit_date()
                return True
            if self.master_date_extraction():
                return True
            if self.verify_date_found():
                return True
            return False
            # if self.soup.tag_body:
            #     if self.date_attempt_one():
            #         return True
            #     if self.date_attempt_two():
            #         return True
            #     if self.date_attempt_three():
            #         return True
            #     if self.date_attempt_four():
            #         return True
            #     if self.date_attempt_five():
            #         return True
            #     if self.date_attempt_six():
            #         return True
            #     return self.date_attempt_last()
        except Exception as e:
            Log.d(f"Date Extraction Failure. Error=[ {e} ]")
            return False

    def master_date_extraction(self):
        try:
            time_tag = self.soup.soup.find("time")
            if time_tag:
                time_attr = Tag.get_attribute(time_tag, "datetime")
                if time_attr and self.attempt_date_parse_set(time_attr):
                    return True
            return self.date_attempt_master_two()
        except Exception as e:
            Log.d(f"Failed to master extract date. Error=[ {e} ]")
            return False

    def verify_date_found(self):
        pdate = DICT.get("published_date", self.data, False)
        ndate = DICT.get("date", self.data, False)
        if pdate or ndate:
            return True
        return False

    def date_attempt_master_two(self):
        # temp_date = Tag.search(self.soup.element_meta, DATE_TAGS)
        # if temp_date:
        #     raw_date = self.get_content(temp_date)
        #     if self.attempt_date_parse_set(raw_date):
        #         return True
        return False

    def date_attempt_last(self):
        raw_str = str(self.soup.tag_time)
        extraction_attemt = ReDate.extract_date(raw_str)
        if extraction_attemt:
            if self.attempt_date_parse_set(extraction_attemt):
                return True
        raw_str = str(self.soup.tag_body)
        extraction_attemt = ReDate.extract_date(raw_str)
        if extraction_attemt:
            if self.attempt_date_parse_set(extraction_attemt):
                return True
        return False

    def date_attempt_one(self):
        temp_date = Tag.search(self.soup.tag_body, ["datetime", "dateCreated"], enableAttributes=True)
        if temp_date:
            if self.attempt_date_parse_set(LIST.get(1, temp_date)):
                return True
        return False

    def date_attempt_two(self):
        temp_date = Tag.search(self.soup.tag_body, ["update-time"], enableAttributes=True)
        temp_text = Tag.get_text(LIST.get(2, temp_date))
        if temp_text:
            if self.attempt_date_parse_set(temp_date):
                return True
        return False

    def date_attempt_three(self):
        temp_date = Tag.search(self.soup.tag_body, ["datetime", "dateCreated"], enableAttributes=True)
        content = Tag.get_attribute(LIST.get(2, temp_date), "content")
        if content:
            if self.attempt_date_parse_set(content):
                return True
        return False

    def date_attempt_four(self):
        temp_date = Tag.search(self.soup.tag_body, ["published-date"], enableAttributes=True)
        content = Tag.get_text(LIST.get(2, temp_date))
        if content:
            if self.attempt_date_parse_set(content):
                return True
        return False

    def date_attempt_five(self):
        """ Loops through ALL Span Elements Tags -> Text """
        for tag in self.soup.element_span:
            text = Tag.get_text(tag)
            if text:
                date = ReDate.extract_date(text)
                if date:
                    self.set_data("date", date)
                    return True
        return False

    def date_attempt_six(self):
        temp_date = Tag.search_tag_deep(self.soup.tag_body, ["datePublished"], enableAttributes=True)
        temp_text = Tag.get_text(LIST.get(2, temp_date))
        if temp_text:
            if self.attempt_date_parse_set(temp_text):
                return True
        return False

    def reddit_date(self):
        timestamp_tag = self.soup.soup.find("a", {"data-testid": "post_timestamp"})
        if timestamp_tag:
            date_ready = DATE.parse_reddit_timestamp_to_datetime(timestamp_tag.text)
            self.set_data("published_date", date_ready)

    def attempt_date_parse_set(self, potential_date):
        try:
            date = parse_date(potential_date)
            if date:
                self.set_data("published_date", date)
                return True
            return False
        except Exception as e:
            Log.e(f"Failed to parse date object. DateObj=[ {potential_date} ]", error=e)
            return False

    # -> Author <- #
    # @Ext.timelimit(3)
    def author(self):
        temp_author = Tag.search_element(self.soup.element_meta, ["og:author", "author"])
        if temp_author:
            author = self.get_content(temp_author)
            self.set_data("author", author)
        else:
            self.set_data("author", UNKNOWN)

    # -> Title <- #
    def title(self):
        try:
            title_tag = self.soup.soup.find('title')
            if title_tag:
                title_text = title_tag.string
                self.set_data("title", title_text)
                return
            text = self.soup.tag_h1.text
            self.set_data("title", text)
            if not text or text == "":
                if not self.get_meta_title():
                    self.set_data("title", UNKNOWN)
        except Exception as e:
            Log.e(f"Unable to get Text from tag_h1. Attempting META Extraction.", error=e)
            if not self.get_meta_title():
                self.set_data("title", UNKNOWN)

    def get_meta_title(self):
        temp_title = Tag.search(self.soup.element_meta, ["og:title", "title"])
        if temp_title:
            self.set_data("title", self.get_content(temp_title))
            return True
        return False

    # -> Description <- #
    def description(self):
        if self.master_description():
            return True
        temp_description = Tag.search(self.soup.element_meta, ["og:description", "description"])
        if temp_description:
            descr = self.get_content(temp_description)
            self.set_data("description", descr)

    def master_description(self):
        description_tag = self.soup.soup.find("meta", {"name": "description"})
        if description_tag:
            desc_text = Tag.get_attribute(description_tag, "content")
            if desc_text:
                self.set_data("description", desc_text)
                return True
        return False

    # -> Body <- #
    def body(self):
        # -> If Reddit
        if self.isReddit:
            self.reddit_body()
        elif self.soup.element_p1:
            body = ""
            for p1_item in self.soup.element_p1:
                body = Utils.combine_args_str(body, "\n", self.get_safe_text(p1_item))
            self.set_data("body", body)

    def reddit_body(self):
        # -> Grab Main Post
        post_content = self.soup.soup.findAll("div", {"data-test-id": "post-content"})
        post = ""
        for item in post_content:
            innerTemp = Soup.safe_findAll(item, "p")
            if innerTemp:
                for innerItem in innerTemp:
                    post = Utils.combine_args_str(post, "\n", innerItem.text)
        # -> Grab Comments
        comment_content = self.soup.soup.findAll("div", {"data-testid": "comment"})
        comments = []
        for item in comment_content:
            innerTemp = Soup.safe_findAll(item, "p")
            comment = ""
            if innerTemp:
                for innerItem in innerTemp:
                    comment = Utils.combine_args_str(comment, "\n", innerItem.text)
                comments.append(comment)
        # -> Form Body
        body = post
        index = 1
        for com in comments:
            body = Utils.combine_args_str(body, f"\n\n -> COMMENT {index}: \n", com, "\n")
            index += 1
        self.set_data("body", body)

    def get_safe_text(self, obj):
        try:
            return obj.text
        except Exception as e:
            Log.e("Failed to get Text", error=e)
            return "False"

    # -> ImgUrl <- #

    def img_url(self):
        if self.isReddit and self.reddit_img_url():
            return True
        if self.master_img_url():
            return True
        if self.soup.element_img:
            attemptOne = self.attempt_img_url_one()
            if attemptOne:
                return True
            attemptTwo = self.attempt_img_url_two("data-src")
            if attemptTwo:
                return True
            attemptThree = self.attempt_img_url_two("src")
            if attemptThree:
                return True
            return False

    def master_img_url(self):
        img_tags = self.soup.tag_body.findAll("img")
        if not img_tags:
            return False
        # -> Find Largest img Width in list of image tags.
        highest_width = 0
        highest_img = None
        for inner_img in img_tags:
            result = Tag.get_attribute(inner_img, "width")
            if result:
                result = str(result).replace("%", "")
                if int(result) > highest_width:
                    highest_width = int(result)
                    highest_img = inner_img
        # -> "src" as Key
        img_url_by_attribute = Tag.get_attribute(highest_img, "src")
        if img_url_by_attribute:
            self.set_data("imgUrl", img_url_by_attribute)
            return True
        # -> Convert to String, Regex for all URLS
        img_url_by_regex = URL.find_urls_in_str(highest_img)
        if img_url_by_regex:
            img_url = LIST.get(0, img_url_by_regex)
            self.set_data("imgUrl", img_url)
            return True
        return False

    def reddit_img_url(self):
        img_tags = self.soup.tag_body.findAll("img")
        if not img_tags:
            return False
        for inner_img in img_tags:
            result = Tag.get_attribute(inner_img, "alt")
            if result:
                if str(result).startswith(self.subReddit):
                    img_url = Tag.get_attribute(inner_img, "src")
                    self.set_data("imgUrl", img_url)
                    return True
        return False


    def attempt_img_url_one(self):
        try:
            img_results = Tag.search(self.soup.element_img, "img", enableName=True)
            src_results = Tag.search_attributes(LIST.get(2, img_results), "src")
            potential_url = LIST.get(1, src_results)
            split_url = potential_url.split(" ")
            is_url = URL.is_url(split_url)
            if LIST.get(0, is_url):
                i_url = LIST.get(1, is_url)
                test = URL.is_valid_url(i_url)
                if test:
                    self.set_data("imgUrl", i_url)
                    return True
            return False
        except Exception as e:
            Log.e("Failed to get img.", error=e)
            return False

    def attempt_img_url_two(self, key):
        try:
            img_results = Tag.search(self.soup.element_img, "img", enableName=True)
            if not img_results:
                return False
            src_results = Tag.get_attribute(LIST.get(2, img_results), key)
            if not src_results:
                return False
            potential_url = URL.extract_data_src_url(str(src_results))
            is_url = URL.is_url(potential_url)
            if LIST.get(0, is_url):
                i_url = LIST.get(1, is_url)
                test = URL.is_valid_url(i_url)
                if test:
                    self.set_data("imgUrl", i_url)
                    return True
            return False
        except Exception as e:
            Log.e("Failed to get img.", error=e)
            return False

    def extract_tags(self, tag):
        if len(list(tag)) == 0:
            return NO_STRINGS
        elements = tag.cssselect(
            tag, A_REL_TAG_SELECTOR)
        if not elements:
            elements = tag.cssselect(
                tag, A_HREF_TAG_SELECTOR)
            if not elements:
                return NO_STRINGS

        tags = []
        for el in elements:
            tag = [i for i in el.itertext()]
            if tag:
                tags.append(tag)
        return set(tags)



from bs4 import BeautifulSoup
from bs4.element import Tag as bsTag
from F.LOG import Log
Log = Log("FairWeb.Core.Soup.Parse()")

# -> Step Two -> Convert Response Object to HTML Object
def to_html_from_response(response):
    return BeautifulSoup(response.text, 'html.parser')

def to_html_from_text(text):
    return BeautifulSoup(text, 'html.parser')

def safe_find(tag, term):
    try:
        return tag.find(term)
    except Exception as e:
        Log.e("Failed to find term.", error=e)

def safe_findAll(tag, term):
    try:
        return tag.findAll(term)
    except Exception as e:
        Log.e("Failed to find term.", error=e)

class Parse:
    response = None
    status = False
    soup = None
    raw_text = None
    tag_body: bsTag = None
    tag_head: bsTag = None
    tag_time: bsTag = None
    element_img = None
    element_p1 = None
    tag_h1: bsTag = None
    element_meta = None
    element_span = None

    def __init__(self, response=None, rawText=None):
        self.response = response
        self.raw_text = rawText
        if rawText:
            if self.to_html_from_raw_text():
                self.extract_elements_and_tags()
        elif self.to_html():
            self.extract_elements_and_tags()

    # -> Step Two -> Convert Response Object to HTML Object
    def to_html(self):
        if self.response:
            self.soup = BeautifulSoup(self.response.text, 'html.parser')
            return True
        else:
            return False

    def to_html_from_raw_text(self):
        if self.raw_text:
            self.soup = BeautifulSoup(self.raw_text, 'html.parser')
            return True
        else:
            return False

    # -> Step Three -> Convert HTML into Element/Tag Objects
    def extract_elements_and_tags(self):
        Log.i(f"Extracting HTML Elements.")
        self.tag_body = safe_find(self.soup, "body")  # -> 99%
        self.tag_head = safe_find(self.soup, "head")  # -> 99%
        self.tag_time = safe_find(self.tag_body, "time")  # ->
        self.element_img = safe_findAll(self.tag_body, "img")
        self.element_p1 = safe_findAll(self.soup, "p")  # ->
        self.tag_h1 = safe_find(self.tag_body, "h1")  # ->
        self.element_meta = safe_findAll(self.tag_head, "meta")
        self.element_span = safe_findAll(self.tag_body, "span")
        self.doTest()
        Log.i("Parsing Finished")

    def doTest(self):
        for tr in self.soup.find_all('tr')[2:]:
            tds = tr.find_all('td')
            print("Nome: %s, Cognome: %s, Email: %s" + "% tds[0].text, tds[1].text, tds[2].text")
        pass



from F import LIST
from F import DICT
from F.LOG import Log
from FNLP import URL
from FNLP.Regex import Re as Regex
Log = Log("FairWEB.Crawler.UrlExtractor")

way_back_machine_url = "https://web.archive.org"

def extract_urls_from_soup(soup, masterUrl, stayWithin=None):
    try:
        if not soup:
            return False
        Log.v("Extracting URLs via Soup.")
        soup_urls = soup.findAll('a', href=True)
        fair_urls = URL.find_urls_in_str(soup.__str__())
        extracted_urls = LIST.flatten(soup_urls, fair_urls)
        # -> Add all URLs to Queue
        full_list = private_clean_extracted_urls(masterUrl, extracted_urls)
        if stayWithin:
            filtered_list = private_filter_out_from_stayWithin(full_list, stayWithin)
            return filtered_list
        return full_list
    except Exception as e:
        Log.e("Failed to extract Urls.", error=e)
        return False


def private_clean_extracted_urls(masterUrl, extracted_urls):
    """
    -> Takes raw urls and extensions and combines them
    :param masterUrl:
    :param extracted_urls:
    :return:
    """
    if not extracted_urls:
        Log.i("No Extracted URLS")
        return
    Log.i("Looping extracted urls.")
    cleaned_urls = []
    try:
        for item in extracted_urls:
            _url = DICT.get("href", item)
            Log.d(f"Looking at URL= [ {_url} ]")
            # Web Archive Site
            if str(_url).startswith("/web/") and _url:
                Log.i(f"Url is in way back machine. Fixing url.. [ {_url} ].")
                _url = str(way_back_machine_url) + str(_url)
            # If url begins with / then it's an ext.
            if str(_url).startswith("/"):
                _url = f"https://{masterUrl}{str(_url)}"
            if str(_url).startswith("//"):
                _url = f"https://www.{str(_url)}"
            if _url and str(_url).startswith("http"):
                cleaned_urls.append(_url)
                continue
        return cleaned_urls
    except Exception as e:
        Log.e(f"Failed to handle URLs. Continuing... error=[ {e} ]")
        return extracted_urls

def private_filter_out_from_stayWithin(url_list, stayWithin):
    try:
        filtered_list = []
        for _url in url_list:
            base_url = URL.get_base_url(_url)
            if Regex.contains(stayWithin, base_url):
                filtered_list.append(_url)
            else:
                Log.v(f"Not inside staywithin [ {_url} ]")
                continue
        return filtered_list
    except Exception as e:
        Log.e("Failed to filter URL's.", error=e)
        return url_list


from FW.FusedDownloader import FusedDownloader
from F.LOG import Log
Log = Log("FWEB.Crawler.ArchiveCrawler_v2")


# -> 2. Download Article
def attempt_download(url, saveToDB=True):
    Log.i("Attempting Archive Download.")
    if saveToDB:
        return FusedDownloader.download_v2(url)
    else:
        return FusedDownloader.download(url)


from setuptools import setup, find_packages
import os

current = os.getcwd()

setup(
    name='FairWEB',
    version='4.1.1',
    description='Full HTML WebPage Downloader, Parser and Data Extractor. Web Crawler Included.',
    url='https://github.com/chazzcoin/fairweb',
    author='ChazzCoin',
    author_email='chazzcoin@gmail.com',
    license='BSD 2-clause',
    packages=find_packages(),
    package_data={
        'Resources': ['*.txt']
    },
    install_requires=['FairNLP>=4.0.2', 'FairMongo>=4.1.4',
                      'beautifulsoup4>=4.9.3', 'newspaper3k>=0.2.8', 'pandas>=1.4.2'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)


import time
import urllib.request
import requests
from FW.Core import Soup
from http.cookiejar import CookieJar as cj
from FW.Core.CoreDownloaders import ArticleDownloader
from F import LIST, DICT

import FairResources
from F import EXT
from F.LOG import Log
Log = Log("FWEB.Core.HttpRequest")

FAIL_ENCODING = 'ISO-8859-1'

HEADERS = {
    "scheme": "https",
    "method": "GET",
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.51 Safari/537.36",
    "accept-language": "en-US",
    "accept-encoding": "gzip, deflate, br",
    "content-type": "text/html",
    "accept": "text/html",
    "referer": "https://www.google.com/",
}
HEADERS_EXT = {
    ":authority": "sec.report",
    ":scheme": "https",
    ":method": "GET",
    ":path": "/CIK/Search/Lockheed+Martin+Corporation+Common+Stock",
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.51 Safari/537.36",
    "accept-language": "en-US,en;q=0.9",
    "accept-encoding": "gzip, deflate, br",
    "content-type": "text/html",
    "accept": "application/signed-exchange;v=b3;q=0.7,*/*;q=0.8",
    "referer": "https://sec.report/Senate-Stock-Disclosures",
    "sec-fetch-mode": "no-cors",
    "sec-fetch-dest": "empty",
    "sec-fetch-site": "same-origin",
    "sec-ch-ua-platform": "macOS",
    "upgrade-insecure-requests": "1",
    "sec-ch-ua": "'Not A;Brand';v='99', 'Chromium';v='99', 'Google Chrome';v='99'"
}
COOKIES = {'from-my': 'browser'}

PROXIES = {
  'http': 'http://10.10.1.10:3128',
  'https': 'http://10.10.1.10:1080',
}


# -> Step One -> Call URL and get Raw HTML back in Response Object.
# @Ext.safe_run
# @EXT.sleep(1)
def get_request(url):
    try:
        HEADERS["user-agent"] = FairResources.get_random_user_agent()
        Log.i("Making [ v1 ] HTTP Request.", v=f"URL= [ {url} ] ")
        response = requests.get(url, headers=HEADERS, timeout=10)
        if response.status_code > 205:
            Log.w(f"Failed to make HTTP Request with URL= [ {url} ] ")
            return False
        else:
            Log.s(f"Successful HTTP Request")
            return True, response
    except Exception as e:
        Log.e("Request Failed.", error=e)
        return False

def get_request_3k_to_html(url, parseToSoup=True):
    time.sleep(1)
    try:
        Log.v("Making Request via requests and parsing html via bs4.")
        html = ArticleDownloader.download_html(url)
        Log.v("Request Made.")
        if parseToSoup:
            Log.v("Parsing into Soup Object.")
            parsed_html = Soup.Parse(rawText=html)
            return parsed_html
        return html
    except Exception as e:
        Log.e(f"Failed to get HTML from URL=[ {url} ]", error=e)
        return False

@EXT.sleep(1)
def get_request_v2(url):
    try:
        HEADERS["user-agent"] = FairResources.get_random_user_agent()
        Log.i("Making [ v2 ] HTTP Request.", v=f"URL= [ {url} ] ")
        response = requests.get(url, **get_request_kwargs())
        if response.status_code > 205:
            Log.w(f"Failed to make HTTP Request with URL= [ {url} ] ")
            return False
        else:
            Log.s(f"Successful HTTP Request")
            return True, response
    except Exception as e:
        Log.e("Request Failed.", error=e)
        return False

# -> Step One -> Call URL and get Raw HTML back in Response Object.
def request_to_html(url):
    Log.i(f"Making Request to URL = [ {url} ]")
    resp = get_request(url)
    if resp:
        response = LIST.get(1, resp, False)
        return to_html(response)
    return False

# -> Step Two -> Convert Response Object to HTML Object
def to_html(response):
    try:
        Log.i(f"Parsing Response Text to HTML Objects.")
        return Soup.Parse(response)
    except Exception as e:
        Log.e("Failed to parse into HTML.", error=e)
        return False

# -> Step One -> Call URL and get Raw HTML back in Response Object.
def request_to_html_v2(url):
    Log.i(f"Making [ v2 ] Request to URL = [ {url} ]")
    resp = get_request_v2(url)
    if resp:
        response = LIST.get(1, resp, False)
        return to_html_v2(response)
    return False

def to_html_v2(response):
    if response.encoding != FAIL_ENCODING:
        # return response as a unicode string
        html = response.text
    else:
        html = response.content
        if 'charset' not in response.headers.get('content-type'):
            encodings = response.utils.get_encodings_from_headers(response.headers)
            if len(encodings) > 0:
                response.encoding = encodings[0]
                html = response.text
    return html or ''

def get_request_kwargs():
    """This Wrapper method exists b/c some values in req_kwargs dict
    are methods which need to be called every time we make a request
    """
    HEADERS["user-agent"] = FairResources.get_random_user_agent()
    timeout = 7
    proxies = {}
    return {
        'headers': HEADERS,
        'cookies': cj(),
        'timeout': timeout,
        'allow_redirects': True,
        'proxies': proxies
    }



def urllib(url):
    import urllib.request
    fp = urllib.request.urlopen(url)
    mybytes = fp.read()
    mystr = mybytes.decode("utf8")
    fp.close()
    return mystr

def get_request_urllib(url):
    try:
        Log.i("Making [ urllib ] HTTP Request.", v=f"URL= [ {url} ] ")
        fp = urllib.request.urlopen(url, timeout=10)
        responseCode = DICT.get("code", fp)
        mybytes = fp.read()
        response = mybytes.decode("utf8")
        fp.close()
        if not response and responseCode and int(responseCode) >= 200:
            Log.w(f"Failed to make HTTP Request with URL= [ {url} ] ")
            return False, None
        else:
            Log.s(f"Successful HTTP Request")
            return True, response
    except Exception as e:
        Log.e("Request Failed.", error=e)
        return False





from queue import Queue as Qu
from F.LOG import Log
from FNLP import URL
from FNLP.Regex import Re as Regex

Log = Log("FairWEB.Queue.UrlQueue")

class FQueue:
    maxSize = 0
    noDuplicates = True
    avoidList = None
    cacheName = None
    mainQueue: Qu
    processedQueue: Qu

    def __init__(self, maxSize=0, noDuplicates=True, avoidList=None):
        self.maxSize = maxSize
        self.noDuplicates = noDuplicates
        self.avoidList = avoidList
        self.mainQueue = Qu(self.maxSize)
        self.processedQueue = Qu()

    def isDup(self, obj):
        if obj in self.mainQueue.queue:
            Log.d(f"{obj} is already in queue!")
            return True
        return False

    def hasBeenProcessed(self, obj):
        if obj in self.processedQueue.queue:
            Log.d(f"{obj} has already been processed!")
            return True
        return False

    def clean(self):
        backup = self.mainQueue.queue
        try:
            temp = self.mainQueue.queue
            set_temp = list(set(temp))
            self.mainQueue = Qu(self.maxSize)
            for item in set_temp:
                self.add(item)
            Log.i("--Queue Cleaned.--")
            return True
        except Exception as e:
            self.mainQueue.queue = backup
            Log.e("Failed to clean queue", e)
            return False

    def add_list(self, list_of_objs):
        if not list_of_objs:
            return False
        for item in list_of_objs:
            self.add(item)

    def add(self, obj):
        if self.mainQueue.full():
            return False
        if self.avoidList:
            result = False
            for item in self.avoidList:
                url = URL.get_base_url(obj)
                if Regex.contains(item, url):
                    return
            if not result:
                self.safe_put(obj)
                return
            else:
                Log.d(f"{obj} avoided from queue!")
                return
        else:
            self.safe_put(obj)

    def safe_put(self, obj):
        if self.hasBeenProcessed(obj):
            return False
        elif self.noDuplicates:
            if not self.isDup(obj=obj):
                self.mainQueue.put(obj)
                Log.d(f"{obj} added to queue!")
                return True
            else:
                return False
        else:
            self.mainQueue.put(obj)
            Log.d(f"{obj} added to queue!")
            return True

    def get(self):
        temp = self.mainQueue.get()
        return temp

    def clear_all(self):
        self.mainQueue = Qu(1)
        self.processedQueue = Qu(1)

    def size(self):
        try:
            return self.mainQueue.qsize()
        except Exception as e:
            Log.e("Failed to get Size.", error=e)
            return 0

    def isEmpty(self):
        try:
            return self.mainQueue.empty()
        except Exception as e:
            Log.e("Failed to see if queue is empty.", error=e)
            return 0

    def isFull(self):
        try:
            return self.mainQueue.full()
        except Exception as e:
            Log.e("Failed to see if queue is full.", error=e)
            return 0



# avoid = ['youtube', 'twitter']
# q = TiffanyQueue(avoidList=avoid)
# item1 = "https://youtube.com"
# item2 = "http://twitter.com"
# item3 = 'johnson'
# q.add(item2)
# q.add(item3)
# print(q.mainQueue.queue)
# q.add(item1)
# print(q.mainQueue.queue)


from F import DICT
from F import DATE
from FNLP.Regex import Re as Regex
from F.LOG import Log
from FM.Jarticle.jProvider import jPro as jpro
Log = Log("FWEB.Core.Validator")

jp = jpro()


def validateAndSave(article: {}, saveToArchive=False, setDateToToday=False) -> bool:
    date = DICT.get("published_date", article)
    body = DICT.get("body", article)
    if not date and setDateToToday and body and not body.startswith("Something went wrong"):
        date = DATE.mongo_date_today_str()
        article["published_date"] = date
    if date and not body.startswith("Something went wrong"):
        Log.s(f"Article Validated. Saving Now. DATE=[ {date} ]")
        if saveToArchive:
            save_article(article)
            return True
    else:
        Log.w("Failed to Validate Article.", warning=f"DATE=[ {date} ]")
        return False

# -> MAIN
def validate_article2(article):
    date = DICT.get("published_date", article, default=False)
    body = DICT.get("body", article, default=False)
    if date and body and not body.startswith("Something went wrong"):
        Log.s(f"Article Validated. DATE=[ {date} ]")
        return True
    else:
        Log.w("Failed to Validate Article.", warning=f"DATE=[ {date} ]")
        return False

def validate_article(article):
    """
        Takes in JsonArticle or DownloadWebPage and converts it into JSON
        -> Returns JSON of Article only. or False.
    """
    if type(article) is not dict:
        article = DICT.get("json", article, default=article)
    date = DICT.get("published_date", article, default=False)
    body = DICT.get("body", article, default=False)
    if not v_date(date):
        title = DICT.get("title", article, default=False)
        if v_title(title) and v_body(body):
            Log.s(f"Article Validated by [ TITLE ] and [ BODY ].")
            newArticle = DICT.replace_key_value(article, "published_date", DATE.mongo_date_today_str())
            return newArticle
    # Date has been validated here.
    if v_body(body) and v_date(date):
        Log.s(f"Article Validated by [ BODY ] and [ DATE=[ {date} ] ]")
        return article
    # Last resort, False
    Log.w("Failed to Validate Article.", warning=f"DATE=[ {date} ]")
    return False

def v_body(body):
    if body and not body.startswith("Something went wrong"):
        if len(body) > 10:
            return True
    return False

def v_title(title):
    if title:
        if Regex.contains_any(["UNKNOWN"], title):
            return False
        elif len(title) > 0:
            return True
    return False

def v_date(date):
    if date and date != "False":
        return True
    return False

def save_article(article):
    return jp.ADD_ARTICLES(article)

def mongo_save(func):
    """ -> DECORATOR <- """
    def wrapper(*args):
        # -> func() should return Article in JSON format.
        article = func(*args)
        if article:
            Log.i("ARTICLE HAS BEEN WRAPPED AND VALIDATED!")
            jp.ADD_ARTICLES(article)
            return article
        return article
    return wrapper

def mongo_update(func):
    """ -> DECORATOR <- """
    def wrapper(*args):
        # -> func() should return Article in JSON format.
        article = func(*args)
        if article:
            Log.i("ARTICLE HAS BEEN WRAPPED AND VALIDATED!")
            jp.ADD_ARTICLES(article)
            return article
        return article
    return wrapper